############################################################################
# pioneer.py - Michael Haimes, Ross Glashan                                #
############################################################################
#    MIT SoaR 1.0 - A Python abstraction layer for MobileRobots            #
#    Pioneer3DX robots, and a simulator to simulate their operation        #
#    in a python environment (for testing)                                 #
#                                                                          #
#    Copyright (C) 2006 Massachusetts Institute of Technology              #
#      Written by Michael Haimes <mhaimes@mit.edu>                         #
#        with code from Ross Glashan <rng@mit.edu>                         #
#                                                                          #
#   This program is free software; you can redistribute it and/or modify   #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation; either version 2 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   This program is distributed in the hope that it will be useful,        #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License along#
#   with this program; if not, write to the Free Software Foundation, Inc.,#
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.            #
############################################################################


from time import *
from os import name
import sys
import getopt
from math import pi
from threading import Thread

from soarapp import error, addToClose, removeFromClose, skip
from soarmath import clip

from simulator import MAX_TRANS, MAX_ROT

try:
	from config import SERIAL_PORT_NAME
except ImportError:
	error("You need to run setup.py in the main directory")

METER_SCALE, RADIAN_SCALE = 0.001, 2*pi/4095.0

class Struct:
	def __init__(self, **entries): self.__dict__.update(entries)
Enum = Struct


# P2OS Connection State
(
	STATE_NO_SYNC, 
	STATE_FIRST_SYNC, 
	STATE_SECOND_SYNC, 
	STATE_READY
) = range(4)

# P2OS Connection Setup commands
(
	CMD_SYNC0,
	CMD_SYNC1,
	CMD_SYNC2
) = range(3)

# P2OS Command Argument Types
ArgType  = Enum(
	ARGINT					= 0x3B,
	ARGNINT					= 0x1B,
	ARGSTR					= 0x2B,
)

# P2OS Commands
ArcosCmd = Enum(
	PULSE					=   0,
	OPEN					=   1,
	CLOSE					=   2,
	POLLING					=   3,
	ENABLE					=   4,
	SETA					=   5,
	SETV					=   6,
	SETO					=   7,
	MOVE					=   8,
	ROTATE					=   9,
	SETRV					=  10,
	VEL					=  11,
	HEAD					=  12,
	DHEAD					=  13,
	SAY					=  15,
	JOYREQUEST				=  17,
	CONFIG					=  18,
	ENCODER					=  19,
	RVEL					=  21,
	DCHEAD					=  22,
	SETRA					=  23,
	SONAR					=  28,
	STOP					=  29,
	DIGOUT					=  30,
	VEL2					=  32,
	GRIPPER					=  33,
	ADSEL					=  35,
	GRIPPERVAL				=  36,
	GRIPREQUEST				=  37,
	IOREQUEST				=  40,
	TTY2					=  42,
	GETAUX					=  43,
	BUMPSTALL				=  44,
	TCM2					=  45,
	JOYDRIVE				=  47,
	SONARCYCLE				=  48,
	HOSTBAUD				=  50,
	AUX1BAUD				=  51,
	AUX2BAUD				=  52,
	AUX3BAUD				=  53,
	E_STOP					=  55,
	M_STALL					=  56,
	GYROREQUEST				=  58,
	LCDWRITE				=  59,
	TTY4					=  60,
	GETAUX3					=  61,
	TTY3					=  66,
	GETAUX2					=  67,
	CHARGE					=  68,
	RESET					= 254,
	MAINTENANCE				= 255,
)


# ActiveRobots P2OS Interface
# Provides a python interface to an ActiveMedia Robot over an RS232 link.
class Pioneer:

	# Initialise connection to pioneer on specified port
	def __init__(self, parent):
		self.parent = parent
		if name == "nt":
			from serialwin32 import Serial
		elif name == "posix":
			from serialposix import Serial
		self.portName = SERIAL_PORT_NAME
		self.sonars = [0,0,0,0,0,0,0,0]
		self.sonarsChanged = [0,0,0,0,0,0,0,0]
		self.connectionState = STATE_NO_SYNC
		self.port = None
		self.lastx, self.lasty = 0,0
		self.pos = (0,0,0)
		self.open(Serial)
		self.cmdEnable(1)
		self.parent.addcontrol(skip, self.update, skip)
		addToClose(self.kill)

	def kill(self):	
		self.cmdEnable(0)
		self.sendPacket([ArcosCmd.CLOSE])

	# Open connection to robot
	def open(self, Serial):
		baudRates = [9600,38400,19200,115200,57600]
		curBaudRate = 0
		self.port = Serial(self.portName,baudRates[0])
		self.port.flushInput()
		self.port.flushOutput()
		numAttempts = 3

		while (self.connectionState != STATE_READY):
			if (self.connectionState == STATE_NO_SYNC):
				self.sendPacket([CMD_SYNC0])
			elif (self.connectionState == STATE_FIRST_SYNC):
				self.sendPacket([CMD_SYNC1])
			elif (self.connectionState == STATE_SECOND_SYNC):
				self.sendPacket([CMD_SYNC2])
			elif (self.connectionState == STATE_READY):
				pass

			try:
				pkt = self.recvPacket()
			except:
				if (self.connectionState == STATE_NO_SYNC and numAttempts >= 0):
					numAttempts -= 1
				else:
					curBaudRate += 1
					if (curBaudRate<len(baudRates)):
						self.port.close()
						self.port = Serial(self.portName,baudRates[curBaudRate])
						print "Changing to baud rate",baudRates[curBaudRate]
						self.port.flushInput()
						self.port.flushOutput()
					else:
						raise Exception("No Robot Found")
				continue

			if (pkt[3]==CMD_SYNC0):
				self.connectionState = STATE_FIRST_SYNC
			elif (pkt[3]==CMD_SYNC1):
				self.connectionState = STATE_SECOND_SYNC
			elif (pkt[3]==CMD_SYNC2):
				self.connectionState = STATE_READY

		botName = ""
		botType = ""
		botSubType = ""
		i=4
		while (pkt[i]):
			botName += chr(pkt[i])
			i += 1
		i+=1
		while (pkt[i]):
			botType += chr(pkt[i])
			i += 1
		i+=1
		while (pkt[i]):
			botSubType += chr(pkt[i])
			i += 1

		self.sendPacket([ArcosCmd.OPEN])
		self.update()
		while (0 in self.sonars) or (0.0 in self.sonars): self.update()

		print "P2OS Interface Ready - connected to", botName, botType, botSubType
	
	def destroy(self):	
		removeFromClose(self.kill)
		self.kill()

	def pose(self):
		return self.pos

	# Move by translating with velocity trans (m/sec), 
	# and rotating with velocity rot (deg/sec)
	def motorOutput(self, v, w):
		self.trans = clip(v, -MAX_TRANS, MAX_TRANS)
		self.rot = clip(w, -MAX_ROT, MAX_ROT)
		self.cmdVel(int(self.trans*1000.0))
		sleep(0.1)
		self.cmdRvel(int(self.rot*180/pi))

	def sonarDistances(self):
		return self.sonars

	# update sonars, odometry and stalled
	def update(self, dummyparameter=0):
		changed = [0,0,0,0,0,0,0,0]
		while (1):
			try:
				self.cmdPulse()
			except Exception:
				pass
			for i in range(len(self.sonars)):
				if (self.sonarsChanged[i]):
					changed[i] = 1
			if (0 not in changed):
				break
	
		# Calculate checksum on P2OS packet (see Pioneer manual)
	def calcChecksum(self,data):
		c = 0
		i = 3
		n = data[2]-2
		while (n>1):
			c += (data[i]<<8) | (data[i+1])
			c = c & 0xFFFF
			n -= 2
			i += 2
		if (n>0):
			c = c ^ data[i]
		return c

	# Send a packet to robot
	def sendPacket(self,data):
		pkt = [0xFA,0xFB,len(data)+2]
		for d in data:
			pkt.append(d)
		pkt.append(0)
		pkt.append(0)
		chk = self.calcChecksum(pkt)
		pkt[len(pkt)-2] = (chk>>8)
		pkt[len(pkt)-1] = (chk&0xFF)
		s = reduce(lambda x,y: x+chr(y),pkt,"")
		self.port.write(s)
		sleep(0.05)

	# Block until packet recieved
	def recvPacket(self):
		timeout = 1.0
		data = [0,0,0]
		while (1):
			cnt = 0
			tstart = time()
			while (time()-tstart)<timeout:
				if (self.port.inWaiting()>0):
					data[2] = ord(self.port.read())
					break
			if (time()-tstart)>timeout:
				raise Exception("Read timeout")
			if (data[0] == 0xFA and data[1] == 0xFB):
				break
			data[0] = data[1]
			data[1] = data[2]

		for d in range(data[2]):
			data.append(ord(self.port.read()))

		crc = self.calcChecksum(data)
		if data[len(data)-1]!=(crc&0xFF) or data[len(data)-2]!=(crc>>8):
			raise Exception("Checksum failure")
		if self.port.inWaiting() > 0: 
			return self.recvPacket()
		else:
			return data

	# Send a packet and receive a SIP response from the robot
	def sipSendReceive(self,data):
		self.sendPacket(data)
		recv = self.recvPacket()
		if (recv[3]&0xF0)==0x30:
			# parse all data
			xpos        = recv[ 4] | (recv[ 5]<<8)
			ypos        = recv[ 6] | (recv[ 7]<<8)
			thpos       = recv[ 8] | (recv[ 9]<<8)
			#lvel        = recv[10] | (recv[11]<<8)
			#rvel        = recv[12] | (recv[13]<<8)
			#battery     = recv[14]
			stallbump   = recv[15] | (recv[16]<<8)
			#control     = recv[17] | (recv[18]<<8)
			#flags       = recv[19] | (recv[20]<<8)
			#compass     = recv[21]
			sonarcount  = recv[22]
			sonars = [-1,-1,-1,-1,-1,-1,-1,-1]
			for i in range(sonarcount):
				num = recv[23+3*i]
				sonars[num] = recv[24+3*i] | (recv[25+3*i]<<8)
			#indx = 23 + 3*sonarcount
			#gripstate   = recv[indx]
			#anport      = recv[indx+1]
			#analog      = recv[indx+2]
			#digin       = recv[indx+3]
			#digout      = recv[indx+4]
			#batteryx10  = recv[indx+5] | (recv[indx+6]<<8)
			#chargestate = recv[indx+7]

			# process
			dx, dy = xpos-self.lastx, ypos-self.lasty
			if dx > 60000:
				dx-=65536
			elif dx < -60000:
				dx+=65536
			if dy > 60000:
				dy-=65536
			elif dy < -60000:
				dy+=65536

			self.pos = (self.pos[0]+dx*METER_SCALE,self.pos[1]+dy*METER_SCALE,thpos*RADIAN_SCALE)
			self.lastx, self.lasty = xpos, ypos
			#self.vel = [lvel,rvel]
			#self.battery = battery/10.0
			self.stall = [( (stallbump&0x0001) == 0x0001 ), ( (stallbump&0x0100) == 0x0100 )]
			self.bump = [( (stallbump&0x00FE) >> 1 ), ( (stallbump&0xFE00) >> 9 )]
			self.stalled = self.stall[0] or self.stall[1] or self.bump[0] or self.bump[1]
			for i in range(len(sonars)):
				if sonars[i]!=-1:
					self.sonars[i] = sonars[i]*METER_SCALE
					self.sonarsChanged[i] = 1
				else:
					self.sonarsChanged[i] = 0

			# print
		        	
			"""
			print "pos     :",self.pos 
			print "stall  :",self.stalled
			print "sonars  :",self.sonars
			print "vel     :",self.vel
			print "battery :",self.battery
			print self.battery
			"""
	def cmdPulse(self):
		self.sipSendReceive([ArcosCmd.PULSE])

	def cmdEnable(self,v):
		self.sendPacket([ArcosCmd.ENABLE,ArgType.ARGINT,v,0])

	def cmdVel(self,v):
		absv = int(abs(v))
		if (v>=0):
			data = [ArcosCmd.VEL,ArgType.ARGINT,absv&0xFF,absv>>8]
		else:
			data = [ArcosCmd.VEL,ArgType.ARGNINT,absv&0xFF,absv>>8]
		self.sendPacket(data)

	def cmdVel2(self,l,r):
		self.sendPacket([ArcosCmd.VEL2,ArgType.ARGINT,int(r),int(l)])

	def cmdRvel(self,rv):
		absrv = abs(rv)
		if (rv>=0):
			data = [ArcosCmd.RVEL,ArgType.ARGINT,absrv&0xFF,absrv>>8]
		else:
			data = [ArcosCmd.RVEL,ArgType.ARGNINT,absrv&0xFF,absrv>>8]
		self.sendPacket(data)

	def cmdSay(self,note,duration):
		data = [ArcosCmd.SAY,ArgType.ARGSTR]
		for d in ("%s,%s" % (note,duration)):
			data.append(ord(d))
		data.append(0)
		self.sendPacket(data)
	
