############################################################################
# simulator.py - Michael Haimes                                            #
############################################################################
#    MIT SoaR 1.0 - A Python abstraction layer for MobileRobots            #
#    Pioneer3DX robots, and a simulator to simulate their operation        #
#    in a python environment (for testing)                                 #
#                                                                          #
#    Copyright (C) 2006 Massachusetts Institute of Technology              #
#      Written by Michael Haimes <mhaimes@mit.edu>                         #
#        with code from Ross Glashan <rng@mit.edu>                         #
#                                                                          #
#   This program is free software; you can redistribute it and/or modify   #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation; either version 2 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   This program is distributed in the hope that it will be useful,        #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License along#
#   with this program; if not, write to the Free Software Foundation, Inc.,#
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.            #
############################################################################

from soarapp import parseFile, AppManager
from soarmath import intersection, angle, angleok, clip

from time import time, sleep
from Tkinter import *
from math import sin
from math import cos
from math import acos
from math import atan2
from math import pi
from random import gauss
from threading import Thread
import os

MAX_FPS=12 #maximum steps to compute per second
DRAW_SONARS=True #draw sonars or not
DRAG_UPDATES_ROBOT_ODOMETRY=False #When the robot is dragged around the simulator, is it's internal odometry updated or not
MAX_SIM_WIDTH=510 #max width of simulator canvas
MAX_SIM_HEIGHT=510 #max height of simulator canvas
SONAR_INFO=[(-0.134, 0.08, pi/2), (-0.118, 0.122, 5*pi/18), (-0.077, 0.156, pi/6), (-0.0266, 0.174, pi/18), (0.0266, 0.174, -pi/18), (0.077, 0.156, -pi/6), (0.118, 0.122, -5*pi/18), (0.134, 0.08, -pi/2)] #entries need to be (x,y,th), x,y are the cartesian coordinate locations of the sonars in meters, and th is the orientation of it with respect to the nose of the robot (CCW is positive)
SONAR_VARIANCE = lambda mean: 0.0017*(mean**3) #an unaccurate approximation of the variance of sonar readings based on distance.. It's better than a single value for variance though, since it is definitely dependent on distance
MAX_SONAR_DIST=5 #mean of maximum time sonar waits before saying "I waited enough time, it's this far or farther"
MAX_TRANS=0.5 #maximum forward/backward velocity of the robot
MAX_ROT=0.5 #maximum rotational velocity of the robot
SONAR_GOT_RESPONSE_COLOR="grey" #color for when the sonar doesn't have to wait until MAX_SONAR_DIST
SONAR_NO_RESPONSE_COLOR="red" #color for when it does
ROBOT_POINTS = [(-0.1905, 0.09525), (-0.0889, 0.180975), (0.0, 0.200025), (0.0889, 0.180975), (0.1905, 0.09525), (0.1905, -0.0635), (0.1651, -0.0635), (0.1651, -0.1651), (0.0889, -0.23495), (-0.0889, -0.23495), (-0.1651, -0.1651), (-0.1651, -0.0635), (-0.1905, -0.0635)]
DRAG_ROT_SPEED = -pi/100 #ratio controlling speed the robot rotates when you right-click to drag


#Don't change unless the laws of math change:
CIRCLE_SCALE_FACTOR = 2**(0.5)
MIN_STEPSIZE = 1.0/MAX_FPS

class Simulator:

    #NOTES: first line must be of the form "dimensions(width, height)"
    #to read a world from a file, simply use World(open("filename").readlines())
    #DON'T change the dimensions on the fly, just make a new world 
    class World:

        def wall(self, start, end):
            self.walls.append((start,end))

        def dimensions(self, width, height):
            self.width = width
            self.height = height
            #borders to make collision detection/sonars easier:
            self.wall((0,0),(0,self.height))
            self.wall((0,0),(self.width,0))
            self.wall((self.width,self.height),(0,self.height))
            self.wall((self.width,self.height),(self.width,0))
	
	def initialRobotLocation(self, x, y):
	    self.init = x,y
	    self.initset = True 
    
        def __init__(self, worldfile):
            self.walls = []
	    self.initset = False
            envin = {}
            envin["dimensions"] = self.dimensions
            envin["wall"] = self.wall
	    envin["initialRobotLoc"] = self.initialRobotLocation
            parseFile(worldfile, envin)
	    if not self.initset:
		self.init = self.width/2.0, self.height/2.0

    def __init__(self, parent, worldfile):
        self.parent = parent
	self.win = AppManager.getWindow("Simulator")
        self.world = self.World(worldfile)
        self.initGlobals()
	self.resetbutton = Button(self.win, text="Reset", command=lambda: self.initGlobals(True))
	self.resetbutton.pack(side="top")
        self.initCanvas()
        self.drawWorld()
        self.drawRobot()
      	self.updateSonars() 
	self.parent.addcontrol(self.startmoving, self.onestep, self.stopmoving)
	AppManager.exitcallbacks.append(self.kill)
 
    def startmoving(self):
        self.stopmoving()
        self.currentthread = self.Stepper()
        self.currentthread.setparams(self)
        self.currentthread.start()
    
    def stopmoving(self):
        try:
            self.currentthread.stop()
        except AttributeError:
            pass

    def destroy(self):
	AppManager.exitcallbacks.remove(self.kill)
	self.kill()
    
    def kill(self):
	self.stopmoving()
        AppManager.killWindow("Simulator")

    def initGlobals(self, reset=False):
        self.v = 0.0
        self.w = 0.0
        self.absx, self.absy = self.world.init
        self.absth = 0.0
        self.odx = 0.0
        self.ody = 0.0
        self.odth = 0.0
	if not reset:
	        if self.world.width > self.world.height:
       		     self.simheight = MAX_SIM_WIDTH*self.world.height/self.world.width
           	     self.simwidth = MAX_SIM_WIDTH
        	else:
            	     self.simwidth = MAX_SIM_HEIGHT*self.world.width/self.world.height
            	     self.simheight = MAX_SIM_HEIGHT
        	self.pxmax = self.world.width
        	self.pxmin = 0
        	self.pymax = self.world.height
        	self.pymin = 0
        	self.cxmax = self.simwidth-5
        	self.cxmin = 5
        	self.cymax = 5
        	self.cymin = self.simheight-5
        	self.pxtocx = (self.cxmax-self.cxmin)/float(self.pxmax-self.pxmin)
        	self.pytocy = (self.cymax-self.cymin)/float(self.pymax-self.pymin)    
        	self.cxtopx = (self.pxmax-self.pxmin)/float(self.cxmax-self.cxmin)    
        	self.cytopy = (self.pymax-self.pymin)/float(self.cymax-self.cymin)    
	else:
		self.drawRobot()
		self.updateSonars()
        self.inrobot = False
        self.stalled = False

    def initCanvas(self):
        self.canvas = Canvas(self.win, width = self.simwidth, height = self.simheight, background = "white")
        self.canvas.bind("<Button-1>", self.left_click_down)
	self.canvas.bind("<B1-Motion>", self.left_click_moved)
        self.canvas.bind("<ButtonRelease-1>", self.left_click_up)
	self.canvas.bind("<Button-3>", self.right_click_down)
	self.canvas.bind("<B3-Motion>", self.right_click_moved)
        self.canvas.bind("<ButtonRelease-3>", self.right_click_up)
        self.canvas.pack(side = "top")
        self.odometrytext = StringVar()
        self.odometrylabel = Label(self.win, textvariable=self.odometrytext)
        self.odometrylabel.pack(side = "top")    


    def PtoC(self, (px, py)):
        cx = (px-self.pxmin)*self.pxtocx+self.cxmin
        cy = (py-self.pymin)*self.pytocy+self.cymin
        return (cx, cy)

    def CtoP(self, (cx, cy)):
        px = (cx-self.cxmin)*self.cxtopx+self.pxmin
        py = (cy-self.cymin)*self.cytopy+self.pymin
        return (px, py)

    def drawWorld(self):
        self.canvas.create_rectangle(0,0,5,self.simheight-5, fill = "grey", outline="grey")
        self.canvas.create_rectangle(0,self.simheight-5,self.simwidth-5,self.simheight, fill="grey", outline="grey")
        self.canvas.create_rectangle(5,0,self.simwidth,5, fill="grey", outline="grey")
        self.canvas.create_rectangle(self.simwidth-5, 5, self.simwidth, self.simheight, fill = "grey", outline="grey")
        self.lines = []
        for wall in self.world.walls:
            start, end = wall
            icx, icy = self.PtoC(start)
            fcx, fcy = self.PtoC(end)
            self.lines.append(self.canvas.create_line(icx, icy, fcx, fcy, fill = "black", width = 2))

    def drawRobot(self):
        cx,cy = self.PtoC((self.absx,self.absy))

        def rotandplacepoint((x, y)):
            r = (x**2+y**2)**(0.5)
            th = atan2(x, y)+self.absth
            return (cx+self.pxtocx*r*cos(th), cy+self.pytocy*r*sin(th))
        points = map(rotandplacepoint, ROBOT_POINTS)
        for item in self.canvas.find_withtag("robot"):
            self.canvas.delete(item)
        self.canvas.create_polygon( \
                              points[0][0], points[0][1],  \
                              points[1][0], points[1][1],  \
                              points[2][0], points[2][1],  \
                              points[3][0], points[3][1],  \
                              points[4][0], points[4][1],  \
                              points[5][0], points[5][1],  \
                              points[6][0], points[6][1],  \
                              points[7][0], points[7][1],  \
                              points[8][0], points[8][1],  \
                              points[9][0], points[9][1],  \
                              points[10][0], points[10][1],\
                              points[11][0], points[11][1],\
                              points[12][0], points[12][1],\
                              tags="robot")
	if os.name == "posix": self.canvas.update_idletasks()
        sonarlines = []
        for entry in SONAR_INFO:
            r = (entry[0]**2+entry[1]**2)**(0.5)
            th = atan2(entry[1], entry[0]) + self.absth - pi/2
            xa, ya = (self.absx+r*cos(th), self.absy+r*sin(th))
            xb, yb = xa+MAX_SONAR_DIST*cos(self.absth+entry[2]), ya+MAX_SONAR_DIST*sin(self.absth+entry[2])
            sonarlines.append(((xa, ya),(xb,yb)))
        self.fullsonarlines = sonarlines
    def odometryString(self):
        return "POSE - X: " + `self.odx`[:7] + ", Y: " + `self.ody`[:7] + ", TH: " + `self.odth`[:7]

    def motorOutput(self, v, w):
        self.v = clip(v, -MAX_TRANS, MAX_TRANS)
        self.w = clip(w, -MAX_ROT, MAX_ROT)
    
    def pose(self):
        return (self.odx, self.ody, self.odth)
   
    def sonarDistances(self):
	return self.storedsonars

    def updateSonars(self):
        sonars = []
        if DRAW_SONARS:
            self.canvas.delete("sonarlines")
        for sonar in self.fullsonarlines:
            mindist = False
            lastint = False
            anglegood = False
            for wall in self.world.walls:
                i = intersection(sonar, wall)
                if i:
                    ix, iy = i
                    sx, sy = sonar[0]
                    dist = ((ix-sx)**2+(iy-sy)**2)**(0.5) 
                    if not mindist or (dist < mindist):
                        anglegood = angleok(abs(pi/2-angle(sonar, wall)))
                        mindist = dist
                        lastint = i
            if not mindist:
		mindist = MAX_SONAR_DIST 
	    if DRAW_SONARS:
                icx, icy = self.PtoC(sonar[0])
               	if lastint: 
			fcx, fcy = self.PtoC(lastint)
		else:	
			fcx, fcy = self.PtoC(sonar[1])
                if anglegood:
                    self.canvas.create_line(icx, icy, fcx, fcy, fill = SONAR_GOT_RESPONSE_COLOR, tag = "sonarlines")
                else:
                    self.canvas.create_line(icx, icy, fcx, fcy, fill = SONAR_NO_RESPONSE_COLOR, tag = "sonarlines")
      		self.canvas.lift("robot")
            sonars.append(gauss(mindist, SONAR_VARIANCE(mindist)))
   	self.storedsonars = sonars
 
    def cache(self):
     	try: 
		self.lastlastpos = self.lastpos 
        except:
		pass
	self.lastpos = self.odx, self.ody, self.odth, self.absx, self.absy, self.absth
    def uncache(self):
        self.odx, self.ody, self.odth, self.absx, self.absy, self.absth = self.lastpos
	try:
		if self.collision():
        		self.odx, self.ody, self.odth, self.absx, self.absy, self.absth = self.lastlastpos
	except:
		pass

    def onestep(self, dt):
        self.cache()
        self.move(dt)
        self.stalled = False
        if self.collision():
            self.uncache()
            self.stalled = True
        self.drawRobot()
	self.updateSonars() 

    class Stepper(Thread):
        def setparams(self, parent):
            self.onestep = parent.onestep
    
        def run(self):
            self.keepgoing = True
            t = time()
            while self.keepgoing:
                dt = time()-t;t = time()
		if dt < MIN_STEPSIZE:
			sleep(MIN_STEPSIZE-dt)
			dt = MIN_STEPSIZE
                self.onestep(dt)
    
        def stop(self):
            self.keepgoing = False

    def collision(self):
        for i in self.canvas.find_withtag("robot"):
            try:
        	ba, bb, bc, bd = self.canvas.bbox(i)
           	for item in self.canvas.find_overlapping(ba, bb, bc, bd):
               		if item in self.lines:
                    		return True
	    except TypeError:
		pass
        return False

    def move(self, dt):
        self.odx+=dt*cos(self.odth)*self.v    
        self.absx+=dt*cos(self.absth)*self.v
        self.ody+=dt*sin(self.odth)*self.v
        self.absy+=dt*sin(self.absth)*self.v
        thstep=dt*self.w
        self.odth+=thstep
        self.absth+=thstep
        self.odth%=2*pi
        self.absth%=2*pi
        self.odometrytext.set(self.odometryString())

    def left_click_down(self, event):
        for item in self.canvas.find_withtag("robot"):
            xn, yn, xx, yx = self.canvas.bbox(item)
            if event.x <= xx and event.x >= xn and event.y <= yx and event.y >= yn:
            	self.inrobot = True    

    def left_click_moved(self, event):
        if self.inrobot:
            self.absx, self.absy = self.CtoP((event.x,event.y))
	    if DRAG_UPDATES_ROBOT_ODOMETRY: self.odx, self.ody = self.CtoP((event.x,event.y))
            self.drawRobot()
	    self.updateSonars()

    def left_click_up(self, event):
        if self.inrobot:
            self.absx, self.absy = self.CtoP((event.x,event.y))
	    if DRAG_UPDATES_ROBOT_ODOMETRY: self.odx, self.ody = self.CtoP((event.x,event.y))
            self.drawRobot()
	    self.updateSonars()
            self.inrobot = False

    def right_click_down(self, event):
        for item in self.canvas.find_withtag("robot"):
            xn, yn, xx, yx = self.canvas.bbox(item)
            if event.x <= xx and event.x >= xn and event.y <= yx and event.y >= yn: 
            	self.inrobot = True
		self.lasteventx, self.lasteventy = event.x, event.y 

    def right_click_moved(self, event):
        if self.inrobot:
	    dx, dy = event.x-self.lasteventx, event.y-self.lasteventy
            self.absth+= (dy+dx)*DRAG_ROT_SPEED
	    if DRAG_UPDATES_ROBOT_ODOMETRY: self.odth+= (dy+dx)*DRAG_ROT_SPEED
            self.drawRobot()
	    self.updateSonars()
	    self.lasteventx, self.lasteventy = event.x, event.y 

    def right_click_up(self, event):
        self.inrobot = False
